import java.util.*;

/**
 * Implements a nondeterministic Turing Machine (TM) with a single, left-bounded tape
 */
public class TuringMachine {
	// Integer constants for head movements
	public static final int HOLD = 0, LEFT = 1, RIGHT = 2;
	
	// Tabular representation of nondeterministic transition matrix
	private Set[][] transitionMatrix;
	private TMTape tape;
	private TMState currentState;
	private int currentPosition;
	private Set states, symbols;
	// Utility data structure to map names to instances for states and symbols
	private Map nameToState, nameToSymbol;
	private int numStates, numSymbols;
	
	// Create a new instance for a TM with maximal number of states and symbols
	public TuringMachine(int maxStates, int maxSymbols) {
		tape = new TMTape();
		states = new HashSet(); symbols = new HashSet();
		nameToState = new HashMap(); nameToSymbol = new HashMap();
		numStates = 0; numSymbols = 0;
		transitionMatrix = new Set[maxStates][maxSymbols];
	}
	
	// Retrieve tape for preparing of input
	public TMTape getTape() {
		return tape;
	}

	// Add a new state to the TM if it does not already exist
	public TMState addState(String name, boolean isStart, boolean isTerminal) {
		TMState s = (TMState) nameToState.get(name);
		if (s == null) {
			s = new TMState(name, isStart, isTerminal, numStates);
			states.add(s);
			nameToState.put(name, s);
			numStates++;
		}
		return s;
	}
	
	// Add a new symbol to the TM if it does not already exist
	public TMSymbol addSymbol(String symbol) {
		TMSymbol s = (TMSymbol) nameToSymbol.get(symbol);
		if (s == null) {
			s = new TMSymbol(symbol, numSymbols);
			symbols.add(s);
			nameToSymbol.put(symbol, s);
			numSymbols++;
		}
		return s;
	}
		
	// Add a new transition to the TM (all referenced states & symbols have to exist already)
	public void addTransition(String fromState, String inputSymbol, 
							  String toState, String outputSymbol, int headDirection) {
		TMState origin = (TMState) nameToState.get(fromState);
		TMState target = (TMState) nameToState.get(toState);
		TMSymbol input = (TMSymbol) nameToSymbol.get(inputSymbol);
		TMSymbol output = (TMSymbol) nameToSymbol.get(outputSymbol);
		TMResult result = new TMResult(target, output, headDirection);
		// Create result set if necessary and add transition for nondeterministic relation
		Set tmp = transitionMatrix[origin.getId()][input.getId()];
		if (tmp == null) {
			tmp = new HashSet();
			transitionMatrix[origin.getId()][input.getId()] = tmp;
		}
		tmp.add(result);
	}
	
	// Initialize TM with start state and given position on the tape
	public void init(int pos) {
		currentPosition = pos;
		Iterator stateIt = states.iterator();
		while (stateIt.hasNext()) {
			TMState s = (TMState) stateIt.next();
			if (s.isStart()) {
				currentState = s;
				break;
			}
		}
	}
	
	// Execute TM after previous initialisation, providing trace output if desired.
	// Returns true if TM reached final state or false otherwise
	public boolean execute(boolean doTrace) {
		List configurations = new Vector();     // Breadth-first expansion of configurations
		Set  oldConfigurations = new HashSet(); // Store expanded configurations
		boolean running = true;
		while (running) {
			if (doTrace) printConfig();
			running = step(configurations, oldConfigurations);
		}
		if (doTrace) printConfig();
		return currentState.isTerminal();
	}
	
	// Perform a singe step of the TM, returns true if the TM may continue or false otherwise
	// Expects a list of configurations to expand & a set of already expanded configurations
	private boolean step(List configurations, Set oldConfigurations) {
		TMConfiguration config = null;
		TMSymbol symbol = tape.read(currentPosition);
		// Determine possible transitions
		Set results = transitionMatrix[currentState.getId()][symbol.getId()];
		if (results != null) {
			// Create a new configuration for every possible transition
			Iterator resultsIt = results.iterator();
			while (resultsIt.hasNext()) {
				// Duplicate current tape and its content
				TMTape newTape = (TMTape) this.getTape().clone();
				// Retrieve result and execute transition locally
				TMResult result = (TMResult) resultsIt.next();
				TMState newState = result.getState();
				newTape.write(currentPosition, result.getSymbol());
				int newPosition = currentPosition;
				switch (result.getHeadDirection()) {
					case LEFT: newPosition--; break;
					case RIGHT: newPosition++; break;
					case HOLD: ; // Do nothing
				}
				TMConfiguration newConfig = new TMConfiguration(newTape, newState, newPosition);
				// Check if current configuration has been reached already
				if (!oldConfigurations.contains(newConfig.toString())) {
					System.out.println("*** Adding configuration " + newConfig);
					oldConfigurations.add(newConfig.toString());
					// Append configuration to current list (breadth-first expansion)
					configurations.add(newConfig);
				} 
			}
		}
		// Retrieve first configuration in list if there are still configurations to try
		try {
			config = (TMConfiguration) configurations.remove(0);
		} catch (IndexOutOfBoundsException ex) {
			System.out.println("*** No more configurations to try");
			return false;
		}
		// Set new configuration of TM, indicating if execution may continue
		tape = config.getTape();
		currentState = config.getState();
		currentPosition = config.getPosition();
		return !currentState.isTerminal();
	}
	
	// Formatted ouput of current configuration
	private void printConfig() {
		for (int i = 0; i < currentPosition; i++)
			System.out.print("  ");
		System.out.print(" " + currentState.getId() + "\n");
		System.out.println(tape);
	}
}
